/*
 * Copyright (c) 2014 Wind River Systems, Inc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */
#include <stdio.h>
#include <stddef.h>
#include <sys/mman.h>
#include <string.h>
#include "vmipc.h"

void ipc_init (void *local, int size)
{
    ipc_t *ipc = (ipc_t *)local;

    ipc->msg_offset = offsetof(ipc_t, data);
    ipc->wr_ptr = ipc->rd_ptr = 0;
    ipc->size = size - ipc->msg_offset;
}

void ipc_show (void *p)
{
    ipc_t *ipc = (ipc_t *)p;

    printf ("\nIPC structure %p\n", p);
    printf ("size=%d\n", ipc->size);
    printf ("offset=%d\n", ipc->msg_offset);
    printf ("wr=%d\n", ipc->wr_ptr);
    printf ("rd=%d\n", ipc->rd_ptr);
}

static void raw_read (ipc_t *ipc_r, ipc_t *ipc_l, unsigned char *dst, int size)
{
    int part_size = 0;
    unsigned char *src;
	
    src = &ipc_r->data[ipc_l->rd_ptr];
    if (size + ipc_l->rd_ptr > ipc_r->size)
    {
        part_size = ipc_r->size - ipc_l->rd_ptr;
        memcpy (dst, src, part_size);

        size -= part_size;
        src = ipc_r->data;
        dst += part_size;
		memcpy (dst, src, size);
		ipc_l->rd_ptr = size;
	}
	else
	{
		memcpy (dst, src, size);
		ipc_l->rd_ptr += size;
	}	
}

int ipc_read (void *local, void *remote, ipc_msg_t *msg)
{
    ipc_t *ipc_l = (ipc_t *)local;
    ipc_t *ipc_r = (ipc_t *)remote;

    if (ipc_l->rd_ptr == ipc_r->wr_ptr)
        return 0;
	
    raw_read (ipc_r, ipc_l, (unsigned char*) msg, sizeof (int));
	
    raw_read (ipc_r, ipc_l, (unsigned char*) &msg->type, msg->size + sizeof (int));

    return msg->size;
}

int ipc_write (void *local, void *remote, ipc_msg_t *msg)
{
    ipc_t *ipc_l = (ipc_t *)local;
    unsigned char *src;
    unsigned char *dst;
    int size;
    int part_size = 0;

    src = (unsigned char*) msg;
    dst = &ipc_l->data[ipc_l->wr_ptr];
    size = msg->size + offsetof (ipc_msg_t, data);

    if (size + ipc_l->wr_ptr > ipc_l->size)
    {
        part_size = ipc_l->size - ipc_l->wr_ptr;
        memcpy (dst, src, part_size);
        size -= part_size;
        src += part_size;
        dst = ipc_l->data;
		memcpy (dst, src, size);
		ipc_l->wr_ptr = size;
    }
	else
	{
		memcpy (dst, src, size);
		ipc_l->wr_ptr += size;
	}
    return size + part_size;
}
